/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define __NO_VERSION__

#include "xfs_locl.h"
#include "xfs_cache.h"
#include "xfs_message.h"
#include "xfs_msg_locl.h"
#include "xdeb.h"
#include "xfs_fs.h"

RCSID("$Id: xfs_message.c,v 1.12 1998/06/04 15:51:33 map Exp $");

int xfs_message_installroot(int fd,
			    struct xfs_message_installroot *message,
			    u_int size)
{
    int error = 0;
    
    XFSDEB(XDEBMSG, ("xfs_message_installroot\n"));
    
    if (xfs[fd].root != 0) {
	printk("XFS PANIC Warning: xfs_message_installroot again\n");
	error = EBUSY;
    } else {
	xfs[fd].root = new_xfs_node(&xfs[fd], &message->node); /* VN_HOLD's */
    }

    return error;
}

int xfs_message_installnode(int fd,
			    struct xfs_message_installnode *message,
			    u_int size)
{
    int error = 0;
    struct xfs_node *n, *dp;
    
    XFSDEB(XDEBMSG, ("xfs_message_installnode\n"));
    
    dp = xfs_node_find(&xfs[fd], &message->parent_handle);
#ifndef LINUX2_1
    if (dp) {
	n = new_xfs_node(&xfs[fd], &message->node); /* VN_HOLD's */
	xfs_cache_add(dp, message->name, strlen(message->name), n);
    } else {
	printk("XFS PANIC Warning: xfs_message_install could not find parent\n");
	error = ENOENT;
    }
#else
    if (dp) {
	struct dentry *dentry;
	struct inode *inode = XNODE_TO_VNODE(dp);
	struct dentry *parent = list_entry(inode->i_dentry.next,
					   struct dentry, d_alias);
	struct qstr sqstr;
	
	XFSDEB(XDEBMSG, ("xfs_message_installnode: fetching new node\n"));
	n = new_xfs_node(&xfs[fd], &message->node); /* VN_HOLD's */
	sqstr.name = message->name;
	sqstr.len = strlen(message->name);
	sqstr.hash = full_name_hash(sqstr.name, sqstr.len);
	dentry = d_alloc(parent, &sqstr);
	XFSDEB(XDEBMSG, ("xfs_message_installnode: allocated new entry: %p\n",
			 dentry));
	
	d_add(dentry, XNODE_TO_VNODE(n));
	XFSDEB(XDEBMSG, ("xfs_message_installnode: done d_add\n"));
    } else {
	printk("XFS PANIC Warning: xfs_message_install could not find parent\n");
	error = ENOENT;
    }
#endif /* LINUX2_1 */

  return error;
}

int xfs_message_installattr(int fd, struct xfs_message_installattr *message, u_int size)
{
    int error = 0;
    struct xfs_node *t;
    int i;
    
    XFSDEB(XDEBMSG, ("xfs_message_installattr\n"));

    t = xfs_node_find(&xfs[fd], &message->node.handle);
    if (t != 0) {
#ifdef LINUX2_1
	struct inode *inode = XNODE_TO_VNODE(t);
	struct dentry *dentry = list_entry(inode->i_dentry.next,
					   struct dentry, d_alias);
	
	XFSDEB(XDEBMSG, ("xfs_message_installattr name:%s\n",
			 dentry->d_name.name));
#endif /* LINUX2_1 */

	t->tokens = message->node.tokens;
	t->attr = message->node.attr;
	memmove(t->id, message->node.id, sizeof(t->id));
	memmove(t->rights, message->node.rights, sizeof(t->rights));
	for (i = 0; i < MAXRIGHTS; i++) {
	    XFSDEB(XDEBMSG, ("rights %d:", t->id[i]));
	    XFSDEB(XDEBMSG, (t->rights[i]&XFS_RIGHT_R?"r":"-"));
	    XFSDEB(XDEBMSG, (t->rights[i]&XFS_RIGHT_W?"w":"-"));
	    XFSDEB(XDEBMSG, (t->rights[i]&XFS_RIGHT_X?"x":"-"));
	    XFSDEB(XDEBMSG, ("\n"));
	}
	t->anonrights = message->node.anonrights;
    } else {
	printk("XFS PANIC Warning: xfs_message_installattr didn't find node!\n");
	error = ENOENT;
    }

    return error;
}

int xfs_message_installdata(int fd,
			    struct xfs_message_installdata *message,
			    u_int size)
{
    struct xfs_node *t;
#ifndef LINUX2_1
    struct inode *vp;
#else
    struct dentry *dentry;
#endif /* LINUX2_1 */
    int error = 0;
    
    XFSDEB(XDEBMSG, ("xfs_message_installdata\n"));
    
    t = xfs_node_find(&xfs[fd], &message->node.handle);
    if (t != 0) {
	char tmp[sizeof(message->cache_handle) + 1];
	XFSDEB(XDEBMSG,("cache_handle = '%c%c%c%c'\n",
			(int) message->cache_handle.data[0],
			(int) message->cache_handle.data[1],
			(int) message->cache_handle.data[2],
			(int) message->cache_handle.data[3]));
	
	memmove (tmp, &message->cache_handle, sizeof(message->cache_handle));
	tmp[4] = '\0';
#ifndef LINUX2_1
	current->fs->pwd->i_count++;
	error = open_namei(tmp, 3, 0, &vp, current->fs->pwd);
#else
	dentry = open_namei(tmp, 3, 0);
#endif /* LINUX2_1 */
#ifndef LINUX2_1
	if (error == 0) {
	    if (DATA_FROM_XNODE(t)) {
		iput(DATA_FROM_XNODE(t));
	    }
	    DATA_FROM_XNODE(t) = vp; /* VOP_LOOKUP does an implicit VN_HOLD? */
	    t->tokens = message->node.tokens;
	    t->attr = message->node.attr;
	    memmove(t->id, message->node.id, sizeof(t->id));
	    memmove(t->rights, message->node.rights, sizeof(t->rights));
	    t->anonrights = message->node.anonrights;
	} else
	    printk("XFS PANIC Warning: xfs_message_installdata failed to lookup cache file = %s, error = %d\n", tmp, error);
#else
	if (!IS_ERR(dentry)) {
	    if (DATA_FROM_XNODE(t)) {
		dput(DATA_FROM_XNODE(t));
	    }
	    DATA_FROM_XNODE(t) = dentry; /* VOP_LOOKUP does an implicit VN_HOLD? */
	    t->tokens = message->node.tokens;
	    t->attr = message->node.attr;
	    memmove(t->id, message->node.id, sizeof(t->id));
	    memmove(t->rights, message->node.rights, sizeof(t->rights));
	    t->anonrights = message->node.anonrights;
	} else {
	    printk("XFS PANIC Warning: xfs_message_installdata failed to lookup cache file = %s, errno: %ld\n", tmp, PTR_ERR(dentry));
	    error = ENOTTY; /* XXXXX */
	}
#endif /* LINUX2_1 */
    } else {
	printk("XFS PANIC Warning: xfs_message_installdata didn't find node!\n");
	error = ENOENT;
    }

    return error;
}

#ifdef LINUX2_1
static void
clear_all_childs (struct inode *inode)
{
    struct list_head *alias;

    alias = inode->i_dentry.next;
    while (alias != &inode->i_dentry) {
	struct dentry *dentry;
	struct list_head *subdirs;
	dentry = list_entry(alias, struct dentry, d_alias);
	if (dentry == NULL) {
	    printk("XFS Panic: dentry in alias list is null\n");
	    break;
	}
	subdirs = dentry->d_subdirs.next;
	while (subdirs != &dentry->d_subdirs) {
	    struct list_head *tmp = subdirs;
	    struct dentry *child = list_entry(tmp, struct dentry, d_child);
	    subdirs = tmp->next;
	    XFSDEB(XDEBMSG, ("clear_all_childs child: %*s\n",
			     child->d_name.len, child->d_name.name));
	    d_drop(child);
	}
	XFSDEB(XDEBMSG, ("clear_all_childs parent: %*s\n",
			 dentry->d_name.len, dentry->d_name.name));
	d_drop(dentry);
	alias = alias->next;
    }
}
#endif /* LINUX2_1 */

int xfs_message_invalidnode(int fd, struct xfs_message_invalidnode *message, u_int size)
{
    int error = 0;
    struct xfs_node *t;
    
    XFSDEB(XDEBMSG, ("xfs_message_invalidnode\n"));
    t = xfs_node_find(&xfs[fd], &message->handle);
    if (t != 0) {
	/* XXX Really need to put back dirty data first. */
	if (DATA_FROM_XNODE(t)) {
#ifndef LINUX2_1
	    iput(DATA_FROM_XNODE(t));
	    DATA_FROM_XNODE(t) = (struct inode *) 0;
#else
	    dput(DATA_FROM_XNODE(t));
	    DATA_FROM_XNODE(t) = NULL;
#endif /* LINUX2_1 */
	}
	XFS_TOKEN_CLEAR(t, ~0,
			XFS_OPEN_MASK | XFS_ATTR_MASK |
			XFS_DATA_MASK | XFS_LOCK_MASK);
#ifndef LINUX2_1
	xfs_cache_delete(t);
#else
      	clear_all_childs(XNODE_TO_VNODE(t));
#endif /* LINUX2_1 */
    } else {
	printk("XFS PANIC Warning: xfs_message_invalidnode didn't find node!\n");
	error = ENOENT;
    }

    return error;
}
