/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define __NO_VERSION__

#include "xfs_locl.h"
#include "xfs_dev.h"
#include <asm/unistd.h>
#include <kafs.h>
#include "xfs_syscalls.h"

RCSID("$Id: xfs_syscalls.c,v 1.16 1998/06/04 15:51:54 map Exp $");

/* 
 * Def pag:
 *  33536 <= g0 <= 34560
 *  32512 <= g1 <= 48896
 */

#define XFS_PAG1_LLIM 33536
#define XFS_PAG1_ULIM 34560
#define XFS_PAG2_LLIM 32512
#define XFS_PAG2_ULIM 48896

static gid_t pag_part_one = XFS_PAG1_LLIM ;
static gid_t pag_part_two = XFS_PAG2_LLIM ;

#ifndef LINUX2_1
static int xfs_get_ngroups(void)
{
    int i;

    for (i = 0 ; current->groups[i] != NOGROUP && i < NGROUPS; i++);
    return i;
}
#endif /* LINUX2_1 */

static int xfs_is_pag(void)
{
    
#ifndef LINUX2_1
    if (current->groups[0] != NOGROUP &&
	current->groups[1] != NOGROUP &&
#else
    if (current->ngroups >= 2 &&
#endif /* LINUX2_1 */
	current->groups[0] >= XFS_PAG1_LLIM &&
	current->groups[0] <= XFS_PAG1_ULIM &&
	current->groups[1] >= XFS_PAG2_LLIM &&
	current->groups[1] <= XFS_PAG2_ULIM)
	
	return 1;
    else
	return 0 ;
}


pag_t xfs_get_pag(void)
{
    if (xfs_is_pag()) {
	
	return (((current->groups[0] << 16) & 0xFFFF0000) |
		((current->groups[1] & 0x0000FFFF))) ;

    } else
	return current->uid;
}

static int
xfs_setpag_call(void)
{
    int i ;

#ifdef DEBUG
    XFSDEB(XDEBSYS, ("xfs_setpag_call groups: "));
#ifndef LINUX2_1
    for (i = 0 ; current->groups[i] != NOGROUP && i < NGROUPS; i++)
#else
    for (i = 0 ; i < current->ngroups; i++)
#endif /* LINUX2_1 */
	XFSDEB(XDEBSYS, ("%d ",current->groups[i]));
    XFSDEB(XDEBSYS, ("\n"));
#endif

    if (!xfs_is_pag()) {
#ifndef LINUX2_1
	int ngroups = xfs_get_ngroups();
#endif /* LINUX2_1 */

	/* Check if it fits */
#ifndef LINUX2_1
	if (ngroups + 2 >= NGROUPS) 
#else
	if (current->ngroups + 2 >= NGROUPS) 
#endif /* LINUX2_1 */
	    return -E2BIG;             /* XXX Hmmm, better error ? */

	/* Copy the groups */
#ifndef LINUX2_1
	for (i = ngroups - 1; i >= 0 ; i--) {
#else
	for (i = current->ngroups - 1; i >= 0 ; i--) {
#endif /* LINUX2_1 */
	    current->groups[i+2] = current->groups[i] ; 
	}
#ifndef LINUX2_1
	ngroups += 2 ;
	if (ngroups < NGROUPS)
	    current->groups[ngroups] = NOGROUP;
#else
	current->ngroups += 2 ;
#endif /* LINUX2_1 */

    }

    current->groups[0] = pag_part_one ;
    current->groups[1] = pag_part_two++ ;

    if (pag_part_two > XFS_PAG2_ULIM) {
	pag_part_one++ ;
	pag_part_two = XFS_PAG2_LLIM ;
    }

    /* Opps, we wraped around... do some smarter stuff ? */
    if (pag_part_one > XFS_PAG1_ULIM) 
	pag_part_one = XFS_PAG1_LLIM ;
	
    return 0;
}

int
handle_xfs_syscall (char *a_pathP,
		    int a_opcode,
		    struct ViceIoctl *a_paramsP,
		    int a_followSymlinks)
{
    switch (a_opcode) {
    case XFSSYSCALL_PROBE:
	return 0x47111147;
    case XFSSYSCALL_SET_DEBUG:
	xfsdeb = XDEBANY;
	return 0;
    case XFSSYSCALL_CLEAR_DEBUG:
	xfsdeb = 0;
	return 0;

    case XFSSYSCALL_SET_DEV_DEBUG:
	xfsdeb |= XDEBDEV;
	return 0;
    case XFSSYSCALL_CLEAR_DEV_DEBUG:
	xfsdeb &= ~XDEBDEV;
	return 0;

    case XFSSYSCALL_SET_DNLC_DEBUG:
	xfsdeb |= XDEBDNLC;
	return 0;
    case XFSSYSCALL_CLEAR_DNLC_DEBUG:
	xfsdeb &= ~XDEBDNLC;
	return 0;

    case XFSSYSCALL_SET_NODE_DEBUG:
	xfsdeb |= XDEBNODE;
	return 0;
    case XFSSYSCALL_CLEAR_NODE_DEBUG:
	xfsdeb &= ~XDEBNODE;
	return 0;

    case XFSSYSCALL_SET_VNOPS_DEBUG:
	xfsdeb |= XDEBVNOPS;
	return 0;
    case XFSSYSCALL_CLEAR_VNOPS_DEBUG:
	xfsdeb &= ~XDEBVNOPS;
	return 0;

    case XFSSYSCALL_SET_VFOPS_DEBUG:
	xfsdeb |= XDEBVFOPS;
	return 0;
    case XFSSYSCALL_CLEAR_VFOPS_DEBUG:
	xfsdeb &= ~XDEBVFOPS;
	return 0;

    case XFSSYSCALL_SET_CACHE_DEBUG:
	xfsdeb |= XDEBCACHE;
	return 0;
    case XFSSYSCALL_CLEAR_CACHE_DEBUG:
	xfsdeb &= ~XDEBCACHE;
	return 0;

    case XFSSYSCALL_SET_SYS_DEBUG:
	xfsdeb |= XDEBSYS;
	return 0;
    case XFSSYSCALL_CLEAR_SYS_DEBUG:
	xfsdeb &= ~XDEBSYS;
	return 0;

    case XFSSYSCALL_SET_READDIR_DEBUG:
	xfsdeb |= XDEBREADDIR;
	return 0;
    case XFSSYSCALL_CLEAR_READDIR_DEBUG:
	xfsdeb &= ~XDEBREADDIR;
	return 0;
    default:
	return ENOSYS;
    }
}

#define MIN(a,b) ((a<b)?a:b)

asmlinkage int
sys_afs (int operation,
	 char *a_pathP,
	 int a_opcode,
	 struct ViceIoctl *a_paramsP,
	 int a_followSymlinks)
{
    int error = 0;
    struct ViceIoctl vice_ioctl; 
    struct xfs_message_pioctl msg;
    struct xfs_message_wakeup_data *msg2;
    
    printk("sys_afs operation: %d a_pathP: %p a_opcode: %d a_paramsP: %p a_followSymlinks: %d\n",operation,a_pathP,a_opcode,a_paramsP,a_followSymlinks);
    
    switch (operation) {
    case AFSCALL_PIOCTL:
	printk("xfs_pioctl\n");
	error = copy_from_user(&vice_ioctl,a_paramsP,sizeof(*a_paramsP));
	if (error)
	    return error;
	if (vice_ioctl.in_size > 2048) {
	    printk("xfs_pioctl_call: got a humongous in packet: opcode: %d",
		   a_opcode);
	    return EINVAL;
	}
	if (vice_ioctl.in_size != 0) {
	    error = copy_from_user(&msg.msg,vice_ioctl.in,vice_ioctl.in_size);
	    if (error)
		return error;
	}
	if (a_pathP != NULL) {
#ifndef LINUX2_1
	    struct inode *i_p;
#endif /* LINUX2_1 */
	    struct xfs_node *xn;
#ifdef LINUX2_1
	    struct dentry *dentry;
#endif /* LINUX2_1 */
	    char *tmp;

	    XFSDEB(XDEBMSG, ("xfs_syscall: looking up: %p\n", a_pathP));

#ifndef LINUX2_1
	    error = getname(a_pathP, &tmp);
#else
	    tmp = getname(a_pathP);
#endif /* LINUX2_1 */

	    XFSDEB(XDEBMSG, ("xfs_syscall: looking up: %s\n", tmp));

#ifndef LINUX2_1
	    error = open_namei(tmp, 1, 0, &i_p, current->fs->pwd);
#else
	    dentry = open_namei(tmp, 1, 0);
#endif /* LINUX2_1 */
	    
#ifndef LINUX2_1
	    if (i_p)
		XFSDEB(XDEBMSG,("xfs_syscall: inode: %p inodenum: %lx\n",
				i_p, i_p->i_ino));
	    if (error) {
		XFSDEB(XDEBMSG, ("xfs_syscall: error during namei: %d\n",
				 error));
		putname(tmp);
		return -EINVAL;
	    }
#else
	    if (dentry)
		XFSDEB(XDEBMSG,("xfs_syscall: inode: %p inodenum: %lx\n",
				dentry->d_inode, dentry->d_inode->i_ino));
	    if (IS_ERR(dentry)) {
		XFSDEB(XDEBMSG, ("xfs_syscall: error during namei: %ld\n",
				 PTR_ERR(error)));
		putname(tmp);
		return -EINVAL;
	    }
#endif /* LINUX2_1 */

		
#ifndef LINUX2_1
	    if (strcmp(i_p->i_sb->s_type->name,"xfs") != 0) {
		XFSDEB(XDEBMSG, ("xfs_syscall: %s not in afs\n", tmp));
		putname(tmp);
		iput(i_p);
		return -EINVAL;
	    }
	    xn = VNODE_TO_XNODE(i_p);
	    msg.handle = xn->handle;
	    putname(tmp);
	    iput(i_p);
#else
	    if (strcmp(dentry->d_inode->i_sb->s_type->name,"xfs") != 0) {
		XFSDEB(XDEBMSG, ("xfs_syscall: %s not in afs\n", tmp));
		putname(tmp);
		dput(dentry);
		return -EINVAL;
	    }
	    xn = VNODE_TO_XNODE(dentry->d_inode);
	    msg.handle = xn->handle;
	    putname(tmp);
	    dput(dentry);
#endif /* LINUX2_1 */
	}

	msg.header.opcode = XFS_MSG_PIOCTL;
	msg.opcode = a_opcode;
	
	msg.insize = vice_ioctl.in_size;
	msg.outsize = vice_ioctl.out_size;
	msg.cred.uid = current->uid;
	msg.cred.pag = xfs_get_pag();
	
	error = xfs_message_rpc(0, &msg.header, sizeof(msg)); /* XXX */
	msg2 = (struct xfs_message_wakeup_data *) &msg;
	if (error == 0)
	    error = msg2->error;
	else
	    error = EINVAL;

	if (error == 0) {
	    error = copy_to_user(vice_ioctl.out, msg2->msg,
				 MIN(msg2->len, vice_ioctl.out_size));
	    if (error)
		return error;
	}

	XFSDEB(XDEBMSG, ("xfs_syscall returns error: %d\n", error));

	return -error;
    case AFSCALL_SETPAG:
	return xfs_setpag_call();
	break;
    case AFSCALL_XFS:
	return handle_xfs_syscall(a_pathP, a_opcode, a_paramsP, a_followSymlinks);
    default:
	printk("xfs_syscalls: unimplemented call\n");
	return -EINVAL;
    }
    
}

typedef int (*sys_call_routine)(int operation,
				char *a_pathP,
				int a_opcode,
				struct ViceIoctl *a_paramsP,
				int a_followSymlinks);

extern sys_call_routine sys_call_table[];

static sys_call_routine old_afs_syscall=NULL;

void
install_afs_syscall (void)
{
  old_afs_syscall = sys_call_table[__NR_afs_syscall];
  sys_call_table[__NR_afs_syscall] = &sys_afs;
}

void
restore_afs_syscall (void)
{
  if (old_afs_syscall) {
    sys_call_table[__NR_afs_syscall] = old_afs_syscall;
    old_afs_syscall = NULL;
  }
}
